﻿using Microsoft.Win32;
using MSTSCLib;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Management;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using WindowsFormsApplication1;

namespace LogonShell
{
	public partial class LogOnForm : Form
	{
		#region P/Invoke
		public enum WTS_CONNECTSTATE_CLASS
		{
			WTSActive,
			WTSConnected,
			WTSConnectQuery,
			WTSShadow,
			WTSDisconnected,
			WTSIdle,
			WTSListen,
			WTSReset,
			WTSDown,
			WTSInit
		}

		public enum WTS_INFO_CLASS
		{
			WTSInitialProgram,
			WTSApplicationName,
			WTSWorkingDirectory,
			WTSOEMId,
			WTSSessionId,
			WTSUserName,
			WTSWinStationName,
			WTSDomainName,
			WTSConnectState,
			WTSClientBuildNumber,
			WTSClientName,
			WTSClientDirectory,
			WTSClientProductId,
			WTSClientHardwareId,
			WTSClientAddress,
			WTSClientDisplay,
			WTSClientProtocolType,
			WTSIdleTime,
			WTSLogonTime,
			WTSIncomingBytes,
			WTSOutgoingBytes,
			WTSIncomingFrames,
			WTSOutgoingFrames,
			WTSClientInfo,
			WTSSessionInfo,
			WTSSessionInfoEx,
			WTSConfigInfo,
			WTSValidationInfo,
			WTSSessionAddressV4,
			WTSIsRemoteSession
		}

		[StructLayout(LayoutKind.Sequential)]
		private struct WTS_SESSION_INFO
		{
			public Int32 SessionID;

			[MarshalAs(UnmanagedType.LPStr)]
			public String pWinStationName;

			public WTS_CONNECTSTATE_CLASS State;
		}

		[DllImport("wtsapi32.dll", SetLastError = true)]
		static extern IntPtr WTSOpenServer(string pServerName);
		[DllImport("wtsapi32.dll", SetLastError = true)]
		static extern int WTSEnumerateSessions(IntPtr hServer, int Reserved, int Version, ref IntPtr ppSessionInfo, ref int pCount);
		[DllImport("wtsapi32.dll", SetLastError = true)]
		static extern bool WTSQuerySessionInformationW(IntPtr hServer, int sessionId, WTS_INFO_CLASS wtsInfoClass, out IntPtr ppBuffer, out int pBytesReturned);
		[DllImport("wtsapi32.dll", SetLastError = true)]
		static extern void WTSFreeMemory(IntPtr value);
		[DllImport("wtsapi32.dll", SetLastError = true)]
		static extern int WTSConnectSession(int LogonId, int TargetLogonId, string pPassword, bool bWait);
		[DllImport("kernel32.dll", SetLastError = true)]
		static extern int WTSGetActiveConsoleSessionId();
		[DllImport("uxtheme.dll", SetLastError = true)]
		static extern int SetWindowTheme(IntPtr window, string app, string theme);
		[DllImport("user32.dll", SetLastError = true)]
		static extern bool ExitWindowsEx(uint uFlags, uint dwReason); [DllImport("advapi32.dll", SetLastError = true, BestFitMapping = false, ThrowOnUnmappableChar = true)]
		[return: MarshalAs(UnmanagedType.Bool)]
		static extern bool LogonUser([MarshalAs(UnmanagedType.LPStr)] string pszUserName, [MarshalAs(UnmanagedType.LPStr)] string pszDomain, [MarshalAs(UnmanagedType.LPStr)] string pszPassword, int dwLogonType, int dwLogonProvider, ref IntPtr phToken);
		[DllImport("user32.dll", SetLastError = true)]
		static extern bool PaintDesktop(IntPtr hdc);

		readonly IntPtr WTS_CURRENT_SERVER_HANDLE = IntPtr.Zero;
		#endregion

		bool loggedIn = false;
		bool optionsShown = false;
		int retryCount = 0;
		private GifImage gifImage = null;
		IniFile settings, Locale;
		System.Windows.Forms.Control UsernameBox;
        SystemMenuManager menuManager;
        public LogOnForm(IniFile settings, IniFile Locale)
		{
			InitializeComponent();
			Load += Form1_Load;
			this.settings = settings;
			this.Locale = Locale;
			gifImage = new GifImage(settings.Read(19));
			gifImage.ReverseAtEnd = Boolean.Parse(settings.Read(20));
			pictureBoxLoading.Image = gifImage.GetNextFrame();
			pictureBoxBanner.Image = Image.FromFile(settings.Read(3));
			if (Boolean.Parse(settings.Read(17)))
				buttonCancel.Enabled = true;
			if (settings.Read(18).ToLower() == "enabled")
				ControlBox = true;
			else if (settings.Read(18).ToLower() == "disabled")
			{
				ControlBox = true;
                this.menuManager = new SystemMenuManager(this, SystemMenuManager.MenuItemState.Greyed);
            }
			LoadLocalisation();
			LoadUsernameBox();
			ShowKeySeqDialog();
		}

		#region RDP Events
		private void axMsRdpClient9NotSafeForScripting1_OnDisconnected(object sender, AxMSTSCLib.IMsTscAxEvents_OnDisconnectedEvent e)
		{
			if (loggedIn)
			{
				timerDisconnect.Enabled = true;
			}
			else
			{
				switch (e.discReason)
				{
					case 2055:
						MessageBox.Show("Logon failed");
						break;
					default:
						MessageBox.Show(e.discReason.ToString());
						break;
				}
			}
		}

		private void axMsRdpClient9NotSafeForScripting1_OnLoginComplete(object sender, EventArgs e)
		{
			loggedIn = true;
			timerLogin.Enabled = true;
		}
		#endregion

		#region Timers
		private void timerDisconnect_Tick(object sender, EventArgs e)
		{
			// Check if we haven't already ran 4 times (12 seconds)
			if (retryCount != 3)
			{
				retryCount++;

				// Retrieve all Windows Terminal Services (aka RDP) Sessions active
				List<string> ret = new List<string>();
				IntPtr ppSessionInfo = IntPtr.Zero;
				IntPtr ppBuffer = IntPtr.Zero;

				int count = 0;
				int retval = WTSEnumerateSessions(WTS_CURRENT_SERVER_HANDLE, 0, 1, ref ppSessionInfo, ref count);
				int dataSize = Marshal.SizeOf(typeof(WTS_SESSION_INFO));

				WTS_SESSION_INFO session = new WTS_SESSION_INFO();
				session.SessionID = -1;

				// Loop over all of them and check if any matches the username but is disconencted
				if (retval != 0)
				{
					for (int i = 0; i < count; i++)
					{
						WTS_SESSION_INFO si = (WTS_SESSION_INFO)Marshal.PtrToStructure(ppSessionInfo, typeof(WTS_SESSION_INFO));
						ppSessionInfo += dataSize;

						WTSQuerySessionInformationW(WTS_CURRENT_SERVER_HANDLE, si.SessionID, WTS_INFO_CLASS.WTSUserName, out ppBuffer, out int stringLength);
						string username = Marshal.PtrToStringAuto(ppBuffer);
						//WTSFreeMemory(ppBuffer);

						if (username == UsernameBox.Text)
							if (si.State == WTS_CONNECTSTATE_CLASS.WTSDisconnected)
								session = si;

						ret.Add(si.SessionID + " " + username + " " + si.State + " " + si.pWinStationName);
					}

					Console.WriteLine(ret.Aggregate((a, b) => $"{a}\n{b}"));
				}
				// If none match the criteria, retry in 3 seconds
				if (session.SessionID == -1)
				{
					return;
				}
				// Otherwise tell WTS to redirect the current session to the new user session and log off from the login account
				else
				{
					int retVal = WTSConnectSession(session.SessionID, WTSGetActiveConsoleSessionId(), textBoxPassword.Text, true);
					if (retVal == 0)
						throw Marshal.GetExceptionForHR(Marshal.GetHRForLastWin32Error());

					timerDisconnect.Enabled = false;
					ExitWindowsEx(4, 0);
					Close();
					return;
				}
			}
			else
			{
				retryCount = 0;
				timerDisconnect.Enabled = false;
				MessageBox.Show(this, "Exceeded maximum RDP disconnection time (12 seconds)");
			}
		}

		private void timerLogin_Tick(object sender, EventArgs e)
		{
			// Check if we haven't already ran 10 times (2.5 seconds)
			if (retryCount != 9)
			{
				retryCount++;

				// Retrieve all Windows Terminal Services (aka RDP) Sessions active
				List<string> ret = new List<string>();
				IntPtr ppSessionInfo = IntPtr.Zero;
				IntPtr ppBuffer = IntPtr.Zero;

				int count = 0;
				int retval = WTSEnumerateSessions(WTS_CURRENT_SERVER_HANDLE, 0, 1, ref ppSessionInfo, ref count);
				int dataSize = Marshal.SizeOf(typeof(WTS_SESSION_INFO));

				int sessionId = -1;

				// Loop over all of them and check if any matches the username
				if (retval != 0)
				{
					for (int i = 0; i < count; i++)
					{
						WTS_SESSION_INFO si = (WTS_SESSION_INFO)Marshal.PtrToStructure(ppSessionInfo, typeof(WTS_SESSION_INFO));
						ppSessionInfo += dataSize;

						WTSQuerySessionInformationW(WTS_CURRENT_SERVER_HANDLE, si.SessionID, WTS_INFO_CLASS.WTSUserName, out ppBuffer, out int stringLength);
						string username = Marshal.PtrToStringAuto(ppBuffer);
						//WTSFreeMemory(ppBuffer);

						if (username == UsernameBox.Text)
							sessionId = si.SessionID;

						ret.Add(si.SessionID + " " + username + " " + si.State + " " + si.pWinStationName);
					}

					Console.WriteLine(ret.Aggregate((a, b) => $"{a}\n{b}"));
					//WTSFreeMemory(ppSessionInfo);
				}
				// If none match the criteria, retry in a quarter of a second
				if (sessionId == -1)
				{
					return;
				}
				// Otherwise tell RDP to disconnect so that the session becomes free for a new connection
				else
				{
					timerLogin.Enabled = false;
					System.Threading.Thread.Sleep(1000);
					axMsRdpClient9NotSafeForScripting1.Disconnect();
				}
			}
			else
			{
				retryCount = 0;
				timerLogin.Enabled = false;
				MessageBox.Show(this, "Exceeded maximum RDP login time (2.5 seconds)");
			}
		}
		#endregion

		#region Form Events
		private void Form1_Load(object sender, EventArgs e)
		{
			SetWindowTheme(Handle, " ", " ");
		}

		private void Form1_Shown(object sender, EventArgs e)
		{
			Location = new Point(Location.X, 166);
			MouseOperations.SetCursorPosition(Location.X + 350, Location.Y + 128);
			MouseOperations.MouseEvent(MouseOperations.MouseEventFlags.LeftDown);
			MouseOperations.MouseEvent(MouseOperations.MouseEventFlags.LeftUp);

		}

		private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            Application.Exit();
        }
		private void Form1_KeyDown(object sender, KeyEventArgs e)
		{
			var keySeqOption = settings.Read(14);
			switch (keySeqOption)
			{
				case "Win":
					if (e.KeyCode == Keys.LWin || e.KeyCode == Keys.RWin)
					{
						labelKeySequence.Visible = false;
						pictureBoxKeySequence.Visible = false;
						ShowLogin();
						this.KeyDown -= this.Form1_KeyDown;
					}
					break;
				case "CtrlAltWin":
					if (e.Control && e.Alt && (e.KeyCode == Keys.LWin || e.KeyCode == Keys.RWin))
					{
						labelKeySequence.Visible = false;
						pictureBoxKeySequence.Visible = false;
						ShowLogin();
						this.KeyDown -= this.Form1_KeyDown;
					}
					break;
				case "CtrlAltEsc":
					if (e.Control && e.Alt && e.KeyCode == Keys.Escape)
					{
						labelKeySequence.Visible = false;
						pictureBoxKeySequence.Visible = false;
						ShowLogin();
						this.KeyDown -= this.Form1_KeyDown;
					}
					break;
				default:
					MessageBox.Show("Configuration Error: unknown key sequence " + keySeqOption);
					break;
			}
		}
		#endregion

		#region Control Events
		private void textBox_KeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Enter)
			{
				buttonOK.PerformClick();
			}
		}

		private void buttonOK_Click(object sender, EventArgs e)
		{
			IntPtr user = new IntPtr();
			bool succes = LogonUser(UsernameBox.Text, Environment.MachineName, textBoxPassword.Text, 3, 0, ref user);
			if (!succes)
			{
                MessageBox.Show(Marshal.GetExceptionForHR(Marshal.GetHRForLastWin32Error()).Message);
                return;
			}

			HideLogin();
            this.Text = Locale.Read(10);
            pictureBoxLoading.Image = Image.FromFile(settings.Read(19));
			labelLoading.Visible = true;
			axMsRdpClient9NotSafeForScripting1.Size = Screen.PrimaryScreen.Bounds.Size;
			axMsRdpClient9NotSafeForScripting1.Server = "127.0.0.2";
			axMsRdpClient9NotSafeForScripting1.Domain = Environment.MachineName;
			axMsRdpClient9NotSafeForScripting1.UserName = UsernameBox.Text;
			axMsRdpClient9NotSafeForScripting1.AdvancedSettings9.ClearTextPassword = textBoxPassword.Text;
			axMsRdpClient9NotSafeForScripting1.AdvancedSettings9.EnableCredSspSupport = true;
			IMsRdpClientNonScriptable5 a = (IMsRdpClientNonScriptable5)axMsRdpClient9NotSafeForScripting1.GetOcx();
			a.AllowPromptingForCredentials = false;
			axMsRdpClient9NotSafeForScripting1.Connect();
			//labelLoading.Text = axMsRdpClient9NotSafeForScripting1.ConnectedStatusText; This either needs constant updating or is just useless.
			//this.Dispose();
		}

		private void buttonCancel_Click(object sender, EventArgs e)
		{
			Application.Exit();
		}

		private void buttonOptions_Click(object sender, EventArgs e)
		{
			optionsShown = !optionsShown;
			buttonShutdown.Visible = optionsShown;
			if (optionsShown)
			{
				buttonOptions.Text = Locale.Read("buttonOptions", "Login") + " < <";
				buttonCancel.Location = buttonOK.Location;
				buttonOK.Location = buttonDummy.Location;
			}
			else
			{
				buttonOptions.Text = Locale.Read("buttonOptions", "Login") + " > >";
				buttonOK.Location = buttonCancel.Location;
				buttonCancel.Location = buttonShutdown.Location;
			}
		}

		private void buttonShutdown_Click(object sender, EventArgs e)
		{
			var ShutdownForm = new ShutDownForm(settings, Locale);
			var result = ShutdownForm.ShowDialog();
			if (result == DialogResult.OK)
            {
                HideLogin();
                this.Text = Locale.Read(10);
                pictureBoxLoading.Image = Image.FromFile(settings.Read(19));
                labelShutdown.Visible = true;
            }
			ShutdownForm.Dispose();
        }
		#endregion

		#region Get User Accounts
		private static void PopulateUserList(ComboBox userComboBox)
		{
			userComboBox.Items.Clear();
			SelectQuery query = new SelectQuery("SELECT * FROM Win32_UserAccount WHERE Disabled=False");
			ManagementObjectSearcher searcher = new ManagementObjectSearcher(query);
			foreach (ManagementObject envVar in searcher.Get())
			{
				userComboBox.Items.Add(envVar["Name"]);
			}
		}
		private static string GetLastUserLoggedOn()
		{
			string location = @"SOFTWARE\Microsoft\Windows\CurrentVersion\Authentication\LogonUI";
			var registryView = Environment.Is64BitOperatingSystem ? RegistryView.Registry64 : RegistryView.Registry32;
			using (var hive = RegistryKey.OpenRemoteBaseKey(RegistryHive.LocalMachine, System.Environment.MachineName, registryView))
			{
				using (var key = hive.OpenSubKey(location))
				{
					var item = key.GetValue("LastLoggedOnUser");
					string itemValue = item == null ? "No Logon Found" : item.ToString().Remove(0, 2); ;
					return itemValue;
				}
			}
		}
		#endregion

		#region Miscellaneous

		private void ShowKeySeqDialog()
		{
			var keySeqOption = settings.Read(14);
			if (keySeqOption.ToLower() != "false" && (keySeqOption == "Win" || keySeqOption == "CtrlAltWin" || keySeqOption == "CtrlAltEsc"))
			{
				HideLogin();
				labelKeySequence.Visible = true;
				pictureBoxKeySequence.Image = Image.FromFile(settings.Read(15));
				pictureBoxKeySequence.Visible = true;
				this.KeyDown += new System.Windows.Forms.KeyEventHandler(this.Form1_KeyDown);
			}
			else if (keySeqOption.ToLower() == "false")
			{
				// Do nothing :)
			}
			else
				MessageBox.Show("Configuration Error: unknown key sequence " + keySeqOption);
		}

		private void HideLogin()
		{
			textBoxPassword.Visible = false;
			comboBoxUsername.Visible = false;
			textBoxUsername.Visible = false;
			var labels = this.Controls.OfType<Label>();
			foreach (var lbl in labels)
				lbl.Visible = false;
			var buttons = this.Controls.OfType<Button>();
			foreach (var btn in buttons)
				btn.Visible = false;
			this.ClientSize = new System.Drawing.Size(409, 132);
		}

		private void ShowLogin()
		{
			this.ClientSize = new System.Drawing.Size(409, 205);
			textBoxPassword.Visible = true;
			UsernameBox.Visible = true;
			labelPassword.Visible = true;
			labelUsername.Visible = true;
			buttonCancel.Visible = true;
			buttonOK.Visible = true;
			buttonOptions.Visible = true;
			//System.Threading.Thread.Sleep(400);
			MouseOperations.SetCursorPosition(Location.X + 350, Location.Y + 128);
			MouseOperations.MouseEvent(MouseOperations.MouseEventFlags.LeftDown);
			MouseOperations.MouseEvent(MouseOperations.MouseEventFlags.LeftUp);
			LoadUsernameBox();
		}

		private void LoadLocalisation()
		{
			this.Text = Locale.Read(5);
			buttonOK.Text = Locale.Read(1);
			buttonCancel.Text = Locale.Read(2);
			buttonOptions.Text = Locale.Read(8) + " > >";
			buttonShutdown.Text = Locale.Read(9);
			labelPassword.Text = Locale.Read(6);
			labelUsername.Text = Locale.Read(7);
			labelLoading.Text = Locale.Read(11);
            labelShutdown.Text = Locale.Read(14);
        }

		private void LoadUsernameBox()
		{
			var usernameOption = settings.Read(16);
			switch (usernameOption)
			{
				case "Simple":
					textBoxUsername.Visible = true;
					comboBoxUsername.Visible = false;
					this.ActiveControl = textBoxUsername;
					textBoxUsername.Focus();
					UsernameBox = textBoxUsername;
					break;
				case "Last":
					textBoxUsername.Visible = true;
					comboBoxUsername.Visible = false;
					comboBoxUsername.Text = GetLastUserLoggedOn();
					UsernameBox = textBoxUsername;
					this.ActiveControl = textBoxPassword;
					textBoxPassword.Focus();
					break;
				case "List":
					textBoxUsername.Visible = false;
					comboBoxUsername.Visible = true;
					PopulateUserList(comboBoxUsername);
					this.ActiveControl = comboBoxUsername;
					comboBoxUsername.Focus();
					UsernameBox = comboBoxUsername;
					break;
				case "ListLast":
					textBoxUsername.Visible = false;
					comboBoxUsername.Visible = true;
					PopulateUserList(comboBoxUsername);
					comboBoxUsername.Text = GetLastUserLoggedOn();
					UsernameBox = comboBoxUsername;
					this.ActiveControl = textBoxPassword;
					textBoxPassword.Focus();
					break;
				default:
					MessageBox.Show("Configuration Error: unknown login type " + usernameOption);
					break;
			}
		}

		#endregion
	}
}
