﻿using System;
using System.Diagnostics;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace LogonShell
{
	public partial class ShutDownForm : Form
	{
		[DllImport("uxtheme.dll", SetLastError = true)]
		static extern int SetWindowTheme(IntPtr window, string app, string theme);

		string[] actionDescriptions =
		{
			@"Ends your session and shuts down Windows so you can safely turn off power.",
			@"Ends your session, shuts down Windows, and starts Windows again.",
			@"Maintains your session, keeping the computer running on low power with data still in memory."
		};

		IniFile settings, Locale;

        public ShutDownForm(IniFile settings, IniFile Locale)
		{
			InitializeComponent();
			this.settings = settings;
			this.Locale = Locale;
			this.Text = Locale.Read(12);
			pictureBoxBanner.Image = Image.FromFile(settings.Read(3));
			pictureBoxLoading.Image = new GifImage(settings.Read(19)).GetNextFrame();
			pictureBoxComputer.Image = Image.FromFile(settings.Read("Icon", "Shutdown")); // TODO: Make the full feature
			comboBox1.Items.Clear();
			comboBox1.Items.Add(Locale.Read(15));
			comboBox1.Items.Add(Locale.Read(16));
			comboBox1.Items.Add(Locale.Read(17));
			actionDescriptions[0] = Locale.Read(18);
			actionDescriptions[1] = Locale.Read(19);
			actionDescriptions[2] = Locale.Read(20);
			labelWdyw.Text = Locale.Read(12);
			buttonOK.Text = Locale.Read(1);
			buttonCancel.Text = Locale.Read(2);
		}

		private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
		{
			if (comboBox1.SelectedIndex != -1)
			{
				labelDescription.Text = actionDescriptions[comboBox1.SelectedIndex];
			}
		}

		private void buttonCancel_Click(object sender, EventArgs e)
		{
			Close();
		}

		private void ShutDownForm_Load(object sender, EventArgs e)
		{
			comboBox1.SelectedIndex = 0;
			SetWindowTheme(Handle, " ", " ");
		}
		private void buttonOK_Click(object sender, EventArgs e)
		{
			ProcessStartInfo psi = null;
			var LoadAnim = true;
			switch (comboBox1.SelectedIndex)
			{
				case 0:
					psi = new ProcessStartInfo("shutdown", "/s /t 0");
					break;
				case 1:
					psi = new ProcessStartInfo("shutdown", "/r /t 0");
					break;
				case 2:
					Application.SetSuspendState(PowerState.Suspend, true, true);
					LoadAnim = false;
					break;
				default:
					MessageBox.Show("The selected option isn't available.");
					LoadAnim = false;
					break;
			}
			if (psi != null)
			{
				psi.CreateNoWindow = true;
				psi.UseShellExecute = false;
				Process.Start(psi);
			}
			if (LoadAnim)
				this.DialogResult = DialogResult.OK;
			this.Close();
			return;
		}
	}
}
